function [x,fval_low,fval_up,exitflag,output] = gosolver(fun,lb,ub,constr,OPTIONS)

%*****************************************************************************
%*                                                                           *
%* GOSOLVER finds a global minimum of a (non-)convex function of several     *
%* variables with respect to (non-)lineare inequality constraints.           *
%* gosolver attempts to solve problems of the form:                          *
%*                                                                           *
%*    min F(X)  subject to:    C(X) <= 0                                     *
%*     X                     LB <= X <= UB                                   *
%*                                                                           *
%* with a (twice) continuously differentiable functions F and                *
%* continuous functions C.                                                   *
%*                                                                           *
%* The input/output-structure is:                                            *
%*                                                                           *
%*  [x,fval,exitflag,output]=gosolver(fun,lb,ub,constr,options)              *
%*                                                                           * 
%* Finite differences and automatic differentiation are not implemented      *
%* yet. The first and/or second derivative(s) of the objective must be       *
%* passed by the user.                                                       *
%*                                                                           *
%* The available branch & bound algorithms are alpha B&B, unimodal form B&B  *
%* and a hybrid B&B algorithm. For the unimodal form B&B one has to pass     *
%* the first derivative of the objective function as second output argument  *
%* of fun. For the other algorithms one has to pass first and second         *
%* derivatives of the objective as second and third output argument.         *
%*                                                                           *
%* The (possibly) not box chaped feasible set of the problem that is given   *
%* by C is relaxed and approximated by boxes.                                *
%*                                                                           *
%*                                                                           *
%* One con choose different branching rules for the algorithms.              *
%* See options.Point and options.Split for more information.                 *
%*                                                                           *
%* INPUT:                                                                    *
%* ------                                                                    *
%*                                                                           *
%*   fun        : objective function                                         *
%*                form: [F,DF,D2F] = fun(x)                                  *
%*                       F: IR^n -> IR,  DF: IR^n -> IR^(n x 1)              *
%*                     D2F: IR^n -> IR^(n x n)                               *
%*                                                                           *
%*   lb         : lower bounds on x                                          *
%*                                                                           *
%*   ub         : upper bounds on x                                          *
%*                                                                           *
%*   constr     : vector of an overall number of m (non-)linear              *
%*                inequality constraints (optional)                          *
%*                form: G = constr(x)                                        *
%*                      G: IR^n -> IR^m                                      *
%*                                                                           *
%*   options    : structure with different options for optimization          *
%*                (optional)                                                 *
%*                MaxIter      : max. number of iterations                   *
%*                               DEFAULT: 1000*n                             *
%*                TolConv      : tolerance for convergence                   *
%*                               DEFAULT: 1e-3                               *
%*                Algori       : algorithm used for optimization             *
%*                               aBB_box   - alpha branch & bound method     *
%*                               ufBB_box  - unimodal form branch & bound    *
%*                                           method                          *
%*                               hybBB_box - hybrid branch & bound method    *
%*                               DEFAULT: aBB_box                            *
%*                                                                           *
%*                Split        : choice of the edge in the branching rule    *
%*                               max     - split the longest edges of a box  *
%*                               gradmax - split the longest edges weighted  *
%*                                         with an interval gradient of the  *
%*                                         objective                         *
%*                               gradmid - split the longest edges weighted  *
%*                                         with an interval gradient of the  *
%*                                         objective and adjusted with the   *
%*                                         barycenter of the box             *
%*                               ratz    - branching rule from D. Ratz       *
%*                               relmax  - split the longest edge measured   *
%*                                         with some kind of relative length *
%*                               DEFAULT: max                                *
%*                                                                           *
%*                Point        : choice of the point in the branching rule   *
%*                               bary - barycenter of the box                *
%*                               min  - minimum of the objective function    *
%*                               DEFAULT: bary                               *
%*                                                                           *
%*                Relax_para   : how and how often relaxation parameters     *
%*                               are computed - for option all and one       *
%*                               the parameters are not computed exactly     *
%*                               but with techniques from interval analysis  *
%*                               all - on each generated box a (set of)      *
%*                                     relaxation parameter(s) is computed   *
%*                               one - use the parameters from the first     *
%*                                     given global box on each generated    *
%*                                     box                                   *
%*                               sub - uses a given function to compute the  *
%*                                     relaxation parameters                 *
%*                                     ->for the alpha B&B method one has    * 
%*                                       construct a function                *
%*                                       alpha_sub_explicit.m that returns   *
%*                                       the largest eigenvalue of D2F on a  *
%*                                       given box [lb,ub] (see below)       *
%*                                     ->for the unimodal form B&B method    * 
%*                                       one has to construct a function     *
%*                                       lip_sub_explicit.m that returns     *
%*                                       lower and upper bounds of DF on a   *
%*                                       given box [lb,ub]. (see below)      *
%*                                     ->for the hybrid B&B method one has   * 
%*                                       construct both mentioned functions  *
%*                               DEFAULT: all                                *
%*                                                                           *
%*                List         : return the list of boxes and related        *
%*                               parameters generated and computed during    *
%*                               optimization                                *
%*                               on  - returns the list in output.List       *
%*                               off - does not return the List              *
%*                               DEFAULT: off                                *
%*                                                                           *
%*                debug        : if you wish additional output during        *
%*                               optimization set options.debug = []         *
%*                                                                           *
%*                                                                           *
%* OUTPUT:                                                                   *
%* -------                                                                   *
%*                                                                           *
%*   x          : last box chosen in optimization process                    *
%*                                                                           *
%*   fval_low   : lower value on the value of the objective function         *
%*                                                                           *
%*   fval_up    : upper bound on the value of the objective function         *
%*                                                                           *
%*                                                                           *
%*   exitflag   : integer identifying the reason the algorithm terminated    *
%*               -9 : an error occured in the input                          *
%*               -3 : increase iter_max                                      *
%*               -2 : boxes become too small                                 *
%*                1 : epsilon optimal solution found                         *
%*                                                                           *
%*   output     : structure containing information about the optimization    *
%*                Algori : algorithm used for optimization                   *
%*                iter   : number of iterations                              *
%*                Time   : CPU time                                          *
%*                Term   : reason the algorithm terminated                   *
%*                List   : list of boxes and related parameters              *
%*                          generated and computed during optimization       *
%*                                                                           *
%* IMPORTANT:                                                                *
%* ----------                                                                *
%*                                                                           *
%* 1.) If (non-)linear inequality constraints are given, then the box        *
%* [lb,ub] must be that large that the feasible set of the problem is        *
%* in the interior if the given box.                                         *
%*                                                                           *
%* 2.) input/output-structure for alpha_sub_explicit.m                       *
%*                                                                           *
%* alpha_sub_explicit.m returns the smallest eigenvalue of D2F on a given    *
%* box [lb,ub]. The input/ouput-structure is:                                *
%*                                                                           *
%*          alpha_relax = alpha_sub_explicit(fname,lb,ub).                   *
%*                                                                           *
%* fname is a function handle of the objective function and lb, ub are the   *
%* lower and upper bounds of a box.                                          *
%*                                                                           *
%* 3.) input/output-structure for lip_sub_explicit.m                         *
%*                                                                           *
%* lip_sub_explicit.m returns lower and upper bounds on DF on a given        *
%* box [lb,ub]. The input/ouput-structure is:                                *
%*                                                                           *
%*          [L_low,L_up] = lip_sub_explicit(fname,lb,ub).                    *
%*                                                                           *
%* fname is a function handle of the objective function and lb, ub are the   *
%* lower and upper bounds of a box.                                          *
%*                                                                           *
%* 4.) If no explicit formula for the relaxation parameters are given        *
%* GOSOLVER needs a package/toolbox for interval arithmetics.                *
%* The toolbox we prefer is IntLab from Prof. Dr. Siegfried M. Rump,         *
%* Institute for Reliable Computing, Hamburg University of Technology.       *
%* GOSOLVER is tested with IntLab, Version 6.                                *
%* If one wishes to use IntLab please add the correct path in line 276.      *
%* NOTE that IntLab is NOT a part of GOSOLVER and that GOSOLVER              *
%* does not affect or imply any license agreement for IntLab.                *
%*                                                                           *
%* The directory should contain a test file called test.m and                *
%* a related example objective and constraint m-files.                       *
%* For the call of the function we refer to test.m                           *
%*                                                                           *
%*****************************************************************************
%*                                                                           *
%* Algorithms/Papers:                                                        *
%* ------------------                                                        *
%*                                                                           *
%* C.S.Adjiman, C.A.Androulakis, C.A.Floudas                                 *
%* A Global Optimization Method, αBB, for general twice-differentiable       *
%* constrained NLPs - I: Theoretical advances                                *
%* Computers and Chemical Engineering, 22 (1998), pp. 1137-1158              *
%*                                                                           *
%* C.S.Adjiman, C.A.Androulakis, C.A.Floudas                                 *
%* A Global Optimization Method, αBB, for general twice-differentiable       *
%* constrained NLPs - II: Implementation and computational results           *
%* Computers and Chemical Engineering, 22 (1998), pp. 1159-1179              *
%*                                                                           *
%* E. Baumann                                                                *
%* Optimal Centered Forms                                                    *
%* BIT Numer. Math., 28 (1988), pp. 80-87                                    *
%*                                                                           *
%* T.Csendes, D.Ratz                                                         *
%* Subdivision direction selection in interval methods for global            *
%* optimization                                                              *
%* SIAM J. Numer. Anal., 34 (1997), pp. 922-938                              *
%*                                                                           *
%* J.Gaerttner                                                               *
%* Unimodale Relaxierungen in der globalen Optimierung: Hybridverafahren     *
%  und numerische Untersuchungen                                             *
%* Bachelorthesis, KIT,                                                      *
%* Institut for Operations Research - Continuous Optimization, 2010.         *
%*                                                                           *
%* R.Horst                                                                   *
%* An Algorithm for Nonconvex Programming Problems                           *
%* Math. Prog., 10 (1976), pp. 312-321                                       *
%*                                                                           *
%* D.Ratz                                                                    *
%* On the Selection of Subdivision Directions in Interval Branch-and-Bound   *
%* Methods for Global Optimization                                           *
%* Journal of Global Optimization, 7 (1995), pp. 183-207                     *
%*                                                                           *
%* S.M.Rump                                                                  *
%* INTLAB - INTerval LABoratory                                              *
%* In Tibor Csendes editor, Developments in Reliable Computing, pp. 77-104   *
%* Kluwer Academic Publishers, Dordrecht, 1999                               *
%*                                                                           *
%*****************************************************************************
%*                                                                           *
%*  This program is free software: you can redistribute it and/or modify     *
%*  it under the terms of the GNU General Public License as published by     *
%*  the Free Software Foundation, either version 3 of the License, or        *
%*  (at your option) any later version.                                      *
%*                                                                           *
%*  This program is distributed in the hope that it will be useful,          *
%*  but WITHOUT ANY WARRANTY; without even the implied warranty of           *
%*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
%*  GNU General Public License for more details.                             *
%*                                                                           *
%*  You should have received a copy of the GNU General Public License        *
%*  along with this program. If not, see <http://www.gnu.org/licenses/>.     *
%*                                                                           *
%*****************************************************************************
%*                                                                           *
%* P.Steuermann - IOR/Karlsruhe Institute of Technology                      *
%* vers.: 0.0.1-0001 - 01/02/2012                                            *
%*                                                                           *
%*****************************************************************************

  disp(' ')
  disp(' ')
  disp('--------------------------------------------------------------------')
  disp('                           GOSOLVER')
  disp('--------------------------------------------------------------------')
  disp('This program is free software: you can redistribute it and/or modify')
  disp('it under the terms of the GNU General Public License as published by')
  disp('the Free Software Foundation, either version 3 of the License, or')
  disp('(at your option) any later version.')
  disp(' ')
  disp('This program is distributed in the hope that it will be useful,')
  disp('but WITHOUT ANY WARRANTY; without even the implied warranty of')
  disp('MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the')
  disp('GNU General Public License for more details.') 
  disp(' ')
  disp('You should have received a copy of the GNU General Public License')
  disp('along with this program.  If not, see <http://www.gnu.org/licenses/>.')
  disp(' ')
  disp(' ')
  disp('--------------------------------------------------------------------')
  disp('--------------------------------------------------------------------')
  disp(' ')
  disp(' ')
  
%% initialize
path_intlab = 'Intlab_V6'; % path to IntLab
                           % adjust the path to your needs

eval(['addpath ',path_intlab]);
startintlab;
intvalinit('SharpIVmult'); % charp (but slow) interval computation

output          = struct;
output.iter     = [];
output.Term     = [];
output.Time     = [];
output.Algori   = [];
output.List     = [];

%% check input and set options

% checkflag input arguments and flag for constraints
checkflag  = 1;
constrflag = 1;

% initialize options fields
options.MaxIter    = [];
options.TolConv    = [];
options.Algori     = [];
options.Split      = [];
options.Point      = [];
options.List       = [];
options.Relax_para = [];

if nargin<5 || isempty(OPTIONS)
    if nargin<3
        % stop if an incorrect number of input arguments is submitted
        disp('Input check failed: Missing input arguments. Please check how gosolver is invoked. Type help gosolver to see more information.');
        checkflag = 0;
    end
else
    % set options for optimization
    field   = fieldnames(OPTIONS);
    n_field = length(field);
    for i=1:n_field
        switch field{i}
            case 'MaxIter'
                options.MaxIter    = OPTIONS.MaxIter;
            case 'TolConv'
                options.TolConv    = OPTIONS.TolConv;
            case 'Algori'
                options.Algori     = OPTIONS.Algori;
            case 'Split'
                options.Split      = OPTIONS.Split;
            case 'Point'
                options.Point      = OPTIONS.Point;
            case 'List'
                options.List       = OPTIONS.List;
            case 'Relax_para'
                options.Relax_para = OPTIONS.Relax_para;
        end
    end
end

%% check box constraints
if isempty(lb) || isempty(ub) || any(isinf(lb)) || any(isinf(ub))
    disp('Input check failed: Infinity bounds are not allowed.');
    checkflag = 0;
else
    [n_l,m_l]=size(lb);
    [n_u,m_u]=size(lb);
    lb = reshape(lb,(n_l*m_l),1);
    ub = reshape(ub,(n_u*m_u),1);
    if (n_l*m_l)~=(n_u*m_u)
        disp('Input check failed:  Check the size of the upper and lower bounds.');
        checkflag = 0;
    end
end
if any(lb>=ub)
    disp('Input check failed: Check upper and lower bounds. Note that fixed variables (i.e. lb=ub) are not allowed.');
    checkflag = 0;
end

% check if (non-)linear inequality constraints are given
if nargin<4 || isempty(constr)
    constrflag = 0;
else
    G = constr(lb);
    if isempty(G)
        constrflag = 0;
    else
        G = constr(infsup(lb,ub));
        if any(inf(G)>0)
            disp('Input check failed: There is no feasible point in [lb,ub].');
            checkflag = 0;
        end
    end
end

%% set solver options if input is submitted correctly and check if all
%% derivatives are available
if checkflag
    % if no options are set use default
    field   = fieldnames(options);
    n_field = length(field);
    for i=1:n_field
        switch field{i}
            case 'MaxIter'
                if isempty(options.MaxIter) || ~isnumeric(options.MaxIter)
                    options.MaxIter = 1000*(n_l*m_l+1);
                end
            case 'TolConv'
                if isempty(options.TolConv) || ~isnumeric(options.TolConv)
                    options.TolConv = 1e-3;
                end
            case 'Algori'
                if isempty(options.Algori)
                    options.Algori  = 'aBB';
                end
            case 'Split'
                if isempty(options.Split)
                    options.Split  = 'max';
                elseif ( ~strncmp(options.Split,'max',3) && ~strncmp(options.Split,'gradmax',7) && ~strncmp(options.Split,'gradmid',7) && ~strncmp(options.Split,'ratz',4) && ~strncmp(options.Split,'relmax',6))
                    disp('Input check failed: Mismatch in branching strategy name. Type help gosolver for available branching strategies.')
                    checkflag = 0;
                end
            case 'Point'
                if isempty(options.Point)
                    options.Point  = 'bary';
                elseif ( ~strncmp(options.Point,'bary',4) && ~strncmp(options.Point,'min',3) )
                    disp('Input check failed: Mismatch in branching point name. Type help gosolver for available branching points.')
                    checkflag = 0;
                end
            case 'List'
                if isempty(options.List) || ( ~strncmp(options.List,'on',4))
                    options.List  = 'off';
                end
            case 'Relax_para'
                if isempty(options.Relax_para)
                    options.Relax_para  = 'all';
                elseif ( ~strncmp(options.Relax_para,'one',3) &&  ~strncmp(options.Relax_para,'all',3) &&  ~strncmp(options.Relax_para,'sub',3))
                    disp('Input check failed: Mismatch in parameter update name. Type help gosolver for available parameter updates.')
                    checkflag = 0;
                end
        end
    end
    
    % check if subfunction for the computation of alpha and L are available
    if strncmp(options.Relax_para,'sub',3)
        if strncmp(options.Algori,'aBB_box',7)
            if ~exist('alpha_sub_explicit','file')
                disp('Input check failed: For the option sub in Relax_para the file alpha_sub_explicit.m for the explizit computation of the relaxation parameter must be given.');
                checkflag = 0;
            end
        elseif strncmp(options.Algori,'ufBB_box',8)
            if ~exist('lip_sub_explicit','file')
                disp('Input check failed: For the option sub in Relax_para the file lip_sub_explicit.m for the explizit computation of the relaxation parameters must be given.');
                checkflag = 0;
            end
        elseif strncmp(options.Algori,'hybBB_box',9)
            if ~exist('alpha_sub_explicit','file') || ~exist('lip_sub_explizit','file')
                disp('Input check failed: For the option sub in Relax_para both files alpha_sub_explicit.m and lip_sub_explicit.m for the explizit computation of the relaxation parameters must be given.');
                checkflag = 0;
            end
        end
    end
     
    % check if derivatives are available and check size
    str_fun          = char(fun);           % as the function may contain 
    str_del          = regexp(str_fun,'@'); % parameters and an @ symbol, 
    str_fun(str_del) = [];                  % one has to delete these symbols, 
    str_del_1        = regexp(str_fun,'('); % parameters, etc. so that the 
    str_del_2        = regexp(str_fun,')'); % function nargout works properly
    n_str_1          = length(str_del_1);
    n_str_2          = length(str_del_2);

    if n_str_1~=n_str_2 % there is problem with the number of brackets in fun
        disp('Input check failed: There is something wrong with the input of the objective function. fun bust be a function handle');
        checkflag = 0;
    else
        if n_str_1>0 % get the function name without any symbols
            for i=n_str_1:(-1):1
                str_fun((str_del_1(i):str_del_2(i))) = [];
            end
        end
        
        % now one can check if derivatives are available - 
        % it is assumed that the second output of fun is the gradient and
        % that the third is the Hessian matrix
        n_out_obj = nargout(str_fun);
        if strncmp(options.Algori,'aBB_box',7) || strncmp(options.Algori,'hybBB_box',9)
            if n_out_obj<3
                disp('Input check failed: Please supply the gradient and the Hessian for the objective function.')
                checkflag = 0;
            else
                [F,DF,D2F]    = fun(0.5*(lb+ub));
                [n_df,m_df]   = size(DF);
                [n_d2f,m_d2f] = size(D2F);
                if m_df~=1
                    disp('Input check failed: Gradient of the objective function must be a column vector.')
                    checkflag = 0;
                elseif n_df~=(n_l*m_l)
                    disp('Input check failed: Check size of the gradient of the objective function.')
                    checkflag = 0;
                elseif n_d2f~=m_d2f
                    disp('Input check failed: Hessian of the objective function must be square.')
                    checkflag = 0;
                elseif n_df~=(n_l*m_l)
                    disp('Input check failed: Check size of the Hessian of the objective function.')
                    checkflag = 0;
                end
            end
        elseif strncmp(options.Algori,'ufBB_box',8)
            if n_out_obj<2
                disp('Input check failed: Please supply the gradient for the objective function.')
                checkflag = 0;
            else
                [F,DF]      = fun(0.5*(lb+ub));
                [n_df,m_df] = size(DF);
                if m_df~=1
                    disp('Input check failed: Gradient of the objective function must be a column vector.')
                    checkflag = 0;
                elseif n_df~=(n_l*m_l)
                    disp('Input check failed: Check size of the gradient of the objective function.')
                    checkflag = 0;
                end
            end
        end
    end
    
    % the Hessian of the objective function must be available for splitting
    % option ratz
    if strncmp(options.Split,'ratz',4)
        if n_out_obj<3
            disp('Input check failed: Please supply the Hessian for the objective function for branching rule "ratz".')
            checkflag = 0;
        end
    end
    
    % if debug is a field in the structure options give some additional output during optimization
    if nargin>4 && isfield(OPTIONS,'debug')
        options.debug    = [];
    end

    %% start optimization
    %% --------------------------------------------------------------------
    %% alphaBB method
    if strncmp(options.Algori,'aBB',7)
        time                                 = cputime;
        if constrflag
            [x,fval_low,fval_up,exitflag,output] = aBB(fun,lb,ub,constr,options);
        else
            [x,fval_low,fval_up,exitflag,output] = aBB_box(fun,lb,ub,options);
        end
        time                                 = cputime-time;
        output.Algori                        = ['alpha BB method with options: ',options.Point,', ',options.Split];
        output.Time                          = time;
    %% --------------------------------------------------------------------
    %% unimodal form BB method
    elseif strncmp(options.Algori,'ufBB',8)
        time                                 = cputime;
        if constrflag
            [x,fval_low,fval_up,exitflag,output] = ufBB(fun,lb,ub,constr,options);
        else
            [x,fval_low,fval_up,exitflag,output] = ufBB_box(fun,lb,ub,options);
        end
        time                                 = cputime-time;
        output.Algori                        = ['unimodal form BB method with options: ',options.Point,', ',options.Split];
        output.Time                          = time;
    %% --------------------------------------------------------------------
    %% quadratic form BB method
    elseif strncmp(options.Algori,'hybBB',9)
        time                                 = cputime;
        if constrflag
            [x,fval_low,fval_up,exitflag,output] = hybBB(fun,lb,ub,constr,options);
        else
            [x,fval_low,fval_up,exitflag,output] = hybBB_box(fun,lb,ub,options);
        end
        time                                 = cputime-time;
        output.Algori                        = ['hybrid BB method with options: ',options.Point,', ',options.Split];
        output.Time                          = time;
    %% --------------------------------------------------------------------
    else
        disp('Input check failed: Mismatch in algorithm name. Type help gosolver for available algorithms.');
        checkflag = 0;
    end
   
end

%% set output if an input error occurred
if ~checkflag
    output.iter     = [];
    output.Term     = 'input error';
    output.Time     = [];
    output.Algori   = [];
    x               = 0.5*(lb+ub);
    fval_low        = [];
    fval_up         = fun(x);
    exitflag        = -9;
end

eval(['rmpath ',path_intlab]);


