function [X_cur,fval_low,fval_up,exitflag,output] = aBB(fname,lb,ub,constr,options)

%**************************************************************************
%*                                                                        *
%* ABB is a branch & bound global optimization solver that uses           *
%* convex relaxations of the objective function to solve problems of      *
%* the form:                                                              *
%*                                                                        *
%*    min F(X)  subject to:    C(X) <= 0                                  *
%*     X                     LB <= X <= UB                                *
%*                                                                        *
%* For more information on the alpha B&B method see:                      *
%*                                                                        *
%* C.S.Adjiman, C.A.Androulakis, C.A.Floudas                              *
%* A Global Optimization Method, αBB, for general twice-differentiable    *
%* constrained NLPs - I: Theoretical advances                             *
%* Computers and Chemical Engineering, 22 (1998), pp. 1137-1158           *
%*                                                                        *
%* C.S.Adjiman, C.A.Androulakis, C.A.Floudas                              *
%* A Global Optimization Method, αBB, for general twice-differentiable    *
%* constrained NLPs - II: Implementation and computational results        *
%* Computers and Chemical Engineering, 22 (1998), pp. 1159-1179           *
%*                                                                        *
%**************************************************************************
%*                                                                        *
%*  This program is free software: you can redistribute it and/or modify  *
%*  it under the terms of the GNU General Public License as published by  *
%*  the Free Software Foundation, either version 3 of the License, or     *
%*  (at your option) any later version.                                   *
%*                                                                        *
%*  This program is distributed in the hope that it will be useful,       *
%*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *
%*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
%*  GNU General Public License for more details.                          *
%*                                                                        *
%*  You should have received a copy of the GNU General Public License     *
%*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *
%*                                                                        *
%**************************************************************************
%*                                                                        *
%* P.Steuermann - IOR/Karlsruhe Institute of Technology                   *
%* vers.: 0.0.1-0001 - 31/01/2012                                         *
%*                                                                        *
%**************************************************************************

debug = 0;
if isfield(options,'debug')
    debug = 1;
end

% set parameters and initialize variables
options_nlp = optimset('Display','off');
tol_conv    = options.TolConv;
split       = options.Split;
split_point = options.Point;
iter_max    = options.MaxIter;
LIST_aBB    = struct;
terminate   = 0;
x_split     = [];
iter        = 1;

time        = cputime;

fval_low    = -inf;
fval_up     = inf;

if strncmp(options.Relax_para,'one',3)
    paraflag = 0;
elseif strncmp(options.Relax_para,'all',3)
    paraflag = 1;
elseif strncmp(options.Relax_para,'sub',3)
    paraflag = 2;
end

% compute relaxation parameter
X_cur       = infsup(lb,ub);
if paraflag <= 1
    [F,DF,D2F]  = fname(X_cur);
    alpha_relax = GG_min(D2F);
else
    alpha_relax = alpha_sub_explicit(fname,lb,ub);
end
LIST_aBB.box    = X_cur;
LIST_aBB.alpha  = alpha_relax;
alpha_relax_old = alpha_relax;

% compute minimum of the relaxation on [lb,ub]
[x_opt,v_opt,nlpflag] = fmincon(@(x)f_alpha(x,fname,alpha_relax,lb,ub),0.5*(lb+ub),[],[],[],[],lb,ub,[],options_nlp);

% if the nlp solver found a KKT point update LIST_aBB and fval_low,fval_up with
% that point, and, otherwise, update only some entries in LIST_aBB with
% "uncritical" values so that LIST_aBB is consistent
if nlpflag>=1
    LIST_aBB.x = x_opt;
    fval_low   = v_opt;
    if all(constr(x_opt)<=0)
        fval_up    = min(fval_up,fname(x_opt));
    end
    LIST_aBB.v = v_opt;
else
    LIST_aBB.x = 0.5*(lb+ub);
    if all(constr(0.5*(lb+ub))<=0)
        fval_up    = min(fval_up,fname(0.5*(lb+ub)));
    end
    LIST_aBB.v = fval_low;
end

% check termination criteria
if fval_up-fval_low<tol_conv
    terminate   = 1;
    exitflag    = 1;
    output.iter = iter;
    output.Term = 'epsilon optimal solution found';
    output.Time = cputime-time;
elseif iter>=iter_max
    terminate   = 1;
    exitflag    = -3;
    output.iter = iter;
    output.Term = 'maximum number of iterations exceeded';
    output.Time = cputime-time;
end

index_min = 1; % there is only one box that can be split in the first iteration 

% set splitting point for options.Point=='min'
if strncmp(split_point,'min',3)
    x_split = LIST_aBB.x(:,1);
end

while ~terminate
    % delete it X-cur and all related data from LIST_aBB
    LIST_aBB.box(:,index_min) = [];
    LIST_aBB.alpha(index_min) = [];
    LIST_aBB.x(:,index_min)   = [];
    LIST_aBB.v(index_min)     = [];

    % split current box
    [X1,X2]      = refine(X_cur,fname,split,split_point,x_split,1e-3);
        
    % check if the intersection of the new boxes and the feasible set is
    % not empty - then compute new parameters and solve relaxed problems 
    % and update LIST_aBB if possible
    % -> first box 
    G = constr(X1);
    if all(inf(G)<=0)
        % compute new parameters alpha_relax on the new boxes X1, X2
        if paraflag == 0
            alpha_relax_1 = alpha_relax;
        elseif paraflag == 1
            [F,DF,D2F]    = fname(X1);
            alpha_relax_1 = GG_min(D2F);
        else
            alpha_relax_1 = alpha_sub_explicit(fname,inf(X1),sup(X1));
        end
    
        % if rounding errors occure in the computation of relaxation parameter
        % then it might increase - we choose the smaller value from the
        % old value and the new value for the relaxation parameter 
        alpha_relax_1 = min(max(alpha_relax_1,alpha_relax_old),0); 
    
        LIST_aBB.box   = [LIST_aBB.box,X1];
        LIST_aBB.alpha = [LIST_aBB.alpha,alpha_relax_1];
    
        LB = inf(X1);
        UB = sup(X1);
        [x_opt,v_opt,nlpflag] = fmincon(@(x)f_alpha(x,fname,alpha_relax_1,LB,UB),0.5*(LB+UB),[],[],[],[],LB,UB,[],options_nlp);
        if nlpflag>=1
            LIST_aBB.x = [LIST_aBB.x,x_opt];
            if all(constr(x_opt)<=0)
                fval_up    = min(fval_up,fname(x_opt));
            end
            LIST_aBB.v = [LIST_aBB.v,v_opt];
        else
            LIST_aBB.x = [LIST_aBB.x,0.5*(lb+ub)];
            if all(constr(0.5*(lb+ub))<=0)
                fval_up    = min(fval_up,fname(0.5*(lb+ub)));
            end
            LIST_aBB.v = [LIST_aBB.v,fval_low];
        end
    end
    
    % -> second box
    G = constr(X2);
    if all(inf(G)<=0)
        % compute new parameters alpha_relax on the new boxes X1, X2
        if paraflag == 0
            alpha_relax_2 = alpha_relax;
        elseif paraflag == 1
            [F,DF,D2F]    = fname(X2);
            alpha_relax_2 = GG_min(D2F);
        else
            alpha_relax_2 = alpha_sub_explicit(fname,inf(X2),sup(X2));
        end
    
        % if rounding errors occure in the computation of relaxation parameter
        % then it might increase - we choose the smaller value from the
        % old value and the new value for the relaxation parameter 
        alpha_relax_2 = min(max(alpha_relax_2,alpha_relax_old),0);
    
        LIST_aBB.box   = [LIST_aBB.box,X2];
        LIST_aBB.alpha = [LIST_aBB.alpha,alpha_relax_2];
        LB = inf(X2);
        UB = sup(X2);
        [x_opt,v_opt,nlpflag] = fmincon(@(x)f_alpha(x,fname,alpha_relax_2,LB,UB),0.5*(LB+UB),[],[],[],[],LB,UB,[],options_nlp);
        if nlpflag>=1
            LIST_aBB.x = [LIST_aBB.x,x_opt];
            if all(constr(x_opt)<=0)
                fval_up    = min(fval_up,fname(x_opt));
            end
            LIST_aBB.v = [LIST_aBB.v,v_opt];
        else
            LIST_aBB.x = [LIST_aBB.x,0.5*(lb+ub)];
            if all(constr(0.5*(lb+ub))<=0)
                fval_up    = min(fval_up,fname(0.5*(lb+ub)));
            end
            LIST_aBB.v = [LIST_aBB.v,fval_low];
        end
    end
    
    % fathorming step
    index_del                 = find(LIST_aBB.v>(fval_up+sqrt(eps)));
    LIST_aBB.box(:,index_del) = [];
    LIST_aBB.x(:,index_del)   = [];
    LIST_aBB.v(index_del)     = [];
    LIST_aBB.alpha(index_del) = [];
    
    % get new box that will be splitted in the next iteration
    [fval_low,index_min] = min(LIST_aBB.v);
    alpha_relax_old      = LIST_aBB.alpha(index_min);
    X_cur                = LIST_aBB.box(:,index_min);
    
    if strncmp(split_point,'min',3)
        x_split = LIST_aBB.x(:,index_min);
    end
    iter = iter+1;
    
    if debug
        disp('upper bound for optimal value')
        fval_up
        disp('lower bound for optimal value')
        fval_low
        disp('smallest and largest alpha in LIST')
        min(LIST_aBB.alpha)
        max(LIST_aBB.alpha)
        disp('number of boxes in LIST')
        length(LIST_aBB.box(1,:))
        pause(2)
    end
    
    % check termination criteria
    if fval_up-fval_low<tol_conv
        terminate   = 1;
        exitflag    = 1;
        output.iter = iter;
        output.Term = 'epsilon optimal solution found';
        output.Time = cputime-time;
        X_cur       = infsup(min(inf(X1),inf(X2)),max(sup(X1),sup(X2)));
    elseif iter>iter_max
        terminate   = 1;
        exitflag    = -3;
        output.iter = iter;
        output.Term = 'maximum number of iterations exceeded';
        output.Time = cputime-time;
    elseif max(sup(X_cur)-inf(X_cur))<1e-6
        terminate   = 1;
        exitflag    = -2;
        output.iter = iter;
        output.Term = 'boxes become too small';
        output.Time = cputime-time;
    end
end

if strncmp(options.List,'on',4)
    output.LIST = LIST_aBB;
end