function [X_cur,fval_low,fval_up,exitflag,output] = hybBB_box(fname,lb,ub,options)

%**************************************************************************
%*                                                                        *
%* HYBBB_BOX is a branch & bound global optimization solver that uses     *
%* convex and unimodal relaxations of the objective function to solve     *
%* problems of the form:                                                  *
%*                                                                        *
%*    min F(X)  subject to:  LB <= X <= UB                                *
%*     X                                                                  *
%*                                                                        *
%* For more information on the unimodal form B&B method see:              *
%*                                                                        *
%* C.S.Adjiman, C.A.Androulakis, C.A.Floudas                              *
%* A Global Optimization Method, αBB, for general twice-differentiable    *
%* constrained NLPs - I: Theoretical advances                             *
%* Computers and Chemical Engineering, 22 (1998), pp. 1137-1158           *
%*                                                                        *
%* C.S.Adjiman, C.A.Androulakis, C.A.Floudas                              *
%* A Global Optimization Method, αBB, for general twice-differentiable    *
%* constrained NLPs - II: Implementation and computational results        *
%* Computers and Chemical Engineering, 22 (1998), pp. 1159-1179           *
%*                                                                        *
%* E. Baumann                                                             *
%* Optimal Centered Forms                                                 *
%* BIT Numer. Math., 28 (1988), pp. 80-87                                 *
%*                                                                        *
%* J.Gaerttner                                                            *
%* Unimodale Relaxierungen in der globalen Optimierung: Hybridverafahren  *
%  und numerische Untersuchungen                                          *
%* Bachelorthesis, KIT,                                                   *
%* Institut for Operations Research - Continuous Optimization, 2010.      *
%*                                                                        *
%**************************************************************************
%*                                                                        *
%*  This program is free software: you can redistribute it and/or modify  *
%*  it under the terms of the GNU General Public License as published by  *
%*  the Free Software Foundation, either version 3 of the License, or     *
%*  (at your option) any later version.                                   *
%*                                                                        *
%*  This program is distributed in the hope that it will be useful,       *
%*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *
%*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
%*  GNU General Public License for more details.                          *
%*                                                                        *
%*  You should have received a copy of the GNU General Public License     *
%*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *
%*                                                                        *
%**************************************************************************
%*                                                                        *
%* P.Steuermann - IOR/Karlsruhe Institute of Technology                   *
%* vers.: 0.0.1-0001 - 31/01/2012                                         *
%*                                                                        *
%**************************************************************************

debug = 0;
if isfield(options,'debug')
    debug = 1;
end

% set parameters and initialize variables
options_nlp = optimset('Display','off');
tol_conv    = options.TolConv;
split       = options.Split;
split_point = options.Point;
iter_max    = options.MaxIter;
LIST_hybBB  = struct;
terminate   = 0;
x_split     = [];
n           = length(lb);
iter        = 1;

time        = cputime;

fval_low    = -inf;
fval_up     = fname(0.5*(lb+ub));

if strncmp(options.Relax_para,'one',3)
    paraflag = 0;
elseif strncmp(options.Relax_para,'all',3)
    paraflag = 1;
elseif strncmp(options.Relax_para,'sub',3)
    paraflag = 2;
end

% compute relaxation parameter
X_cur       = infsup(lb,ub);
if paraflag <= 1
    [F,DF,D2F]  = fname(X_cur);
    alpha_relax = GG_min(D2F);
    L_low       = min(zeros(n,1),inf(DF));
    L_up        = max(zeros(n,1),sup(DF));
else
    alpha_relax  = alpha_sub_explicit(fname,lb,ub);
    [L_low,L_up] = lip_sub_explicit(fname,lb,ub);
    L_low        = min(zeros(n,1),L_low);
    L_up         = max(zeros(n,1),L_up);
end

LIST_hybBB.box   = X_cur;
LIST_hybBB.alpha = alpha_relax;
LIST_hybBB.L.low = L_low;
LIST_hybBB.L.up  = L_up;
alpha_relax_old  = alpha_relax;
L_low_old        = L_low;
L_up_old         = L_up;

% compute optimal center for the unimodal form on [lb,ub]
if all(L_up-L_low>sqrt(eps))
    c = (L_up.*lb-L_low.*ub)./(L_up-L_low);
else
    index_n0    = find(L_up-L_low>sqrt(eps));
    c           = 0.5*(lb+ub);
    c(index_n0) = (L_up(index_n0).*lb(index_n0)-L_low(index_n0).*ub(index_n0))./(L_up(index_n0)-L_low(index_n0));
end

% compute the maximum separation distance of both relaxations 
% (alpha and unimodal) to the original function and compute minimum of 
% the relaxation on [lb,ub] with the smallest maximum separation distance
if 0.125*alpha_relax*norm(ub-lb)^2<(L_low'*(c-ub))
    [x_opt,v_opt,nlpflag] = fmincon(@(x)f_alpha(x,fname,alpha_relax,lb,ub),0.5*(lb+ub),[],[],[],[],lb,ub,[],options_nlp);
else
    nlpflag = 1;
    x_opt   = c;
    v_opt   = fname(c)+L_low'*(ub-c);
end

if nlpflag<1
    x_opt   = c;
    v_opt   = fname(c)+L_low'*(ub-c);
end

LIST_hybBB.x = x_opt;
fval_low     = v_opt;
fval_up      = min(fval_up,fname(x_opt));
LIST_hybBB.v = v_opt;

% check termination criteria
if fval_up-fval_low<tol_conv
    terminate   = 1;
    exitflag    = 1;
    output.iter = iter;
    output.Term = 'epsilon optimal solution found';
    output.Time = cputime-time;
elseif iter>=iter_max
    terminate   = 1;
    exitflag    = -3;
    output.iter = iter;
    output.Term = 'maximum number of iterations exceeded';
    output.Time = cputime-time;
end

index_min = 1; % there is only one box that can be split in the first iteration 

% set splitting point for options.Point=='min'
if strncmp(split_point,'min',3)
    x_split = LIST_hybBB.x(:,1);
end

while ~terminate
    
    % delete it X-cur and all related data from LIST_hybBB
    LIST_hybBB.box(:,index_min)   = [];
    LIST_hybBB.alpha(index_min)   = [];
    LIST_hybBB.x(:,index_min)     = [];
    LIST_hybBB.v(index_min)       = [];
    LIST_hybBB.L.low(:,index_min) = [];
    LIST_hybBB.L.up(:,index_min)  = [];


    % split current box and add the two new boxes to LIST
    [X1,X2]        = refine(X_cur,fname,split,split_point,x_split,1e-3);
    LIST_hybBB.box = [LIST_hybBB.box,X1,X2];
    
    % compute new parameters alpha_relax on the new boxes X1, X2
    if paraflag == 0
        alpha_relax_1 = alpha_relax;
        alpha_relax_2 = alpha_relax;
        L_low_1       = L_low;
        L_low_2       = L_low;
        L_up_1        = L_up;
        L_up_2        = L_up;
    elseif paraflag == 1
        [F,DF,D2F]    = fname(X1);
        alpha_relax_1 = GG_min(D2F);
        L_low_1       = min(zeros(n,1),inf(DF));
        L_up_1        = max(zeros(n,1),sup(DF));
        [F,DF,D2F]    = fname(X2);
        alpha_relax_2 = GG_min(D2F);
        L_low_2       = min(zeros(n,1),inf(DF));
        L_up_2        = max(zeros(n,1),sup(DF));
    else
        alpha_relax_1 = alpha_sub_explicit(fname,inf(X1),sup(X1));
        alpha_relax_2 = alpha_sub_explicit(fname,inf(X2),sup(X2));
        [L_low,L_up]  = lip_sub_explicit(fname,inf(X1),sup(X1));
        L_low_1       = min(zeros(n,1),L_low);
        L_up_1        = max(zeros(n,1),L_up);
        [L_low,L_up]  = lip_sub_explicit(fname,inf(X2),sup(X2));
        L_low_2       = min(zeros(n,1),L_low);
        L_up_2        = max(zeros(n,1),L_up);
    end
    
    % if rounding errors occure in the computation of relaxation parameter
    % then it might increase - we choose the smaller value from the
    % old value and the new value for the relaxation parameter 
    alpha_relax_1 = min(max(alpha_relax_1,alpha_relax_old),0); 
    alpha_relax_2 = min(max(alpha_relax_2,alpha_relax_old),0);
    L_low_1       = max(L_low_1,L_low_old);
    L_up_1        = min(L_up_1,L_up_old);
    L_low_2       = max(L_low_2,L_low_old);
    L_up_2        = min(L_up_2,L_up_old);
    
    LIST_hybBB.alpha = [LIST_hybBB.alpha,alpha_relax_1,alpha_relax_2];
    LIST_hybBB.L.low = [LIST_hybBB.L.low,L_low_1,L_low_2];
    LIST_hybBB.L.up  = [LIST_hybBB.L.up,L_up_1,L_up_2];
    
    
    % compute the maximum separation distance of both relaxations 
    % (alpha and unimodal) to the original function and compute minimum of 
    % the relaxation on [lb,ub] with the smallest maximum separation distance
    
    % -> first box 
    LB = inf(X1);
    UB = sup(X1);
    
    if all(L_up_1-L_low_1>sqrt(eps))
        c_1 = (L_up_1.*LB-L_low_1.*UB)./(L_up_1-L_low_1);
    else
        index_n0    = find(L_up_1-L_low_1>sqrt(eps));
        c_1           = 0.5*(LB+UB);
        c_1(index_n0) = (L_up_1(index_n0).*LB(index_n0)-L_low_1(index_n0).*UB(index_n0))./(L_up_1(index_n0)-L_low_1(index_n0));
    end

    if 0.125*alpha_relax_1*norm(UB-LB)^2<(L_low_1'*(c_1-UB))
        [x_opt,v_opt,nlpflag] = fmincon(@(x)f_alpha(x,fname,alpha_relax_1,LB,UB),0.5*(LB+UB),[],[],[],[],LB,UB,[],options_nlp);
    else
        nlpflag = 1;
        x_opt   = c_1;
        v_opt   = fname(c_1)+L_low_1'*(UB-c_1);
    end

    if nlpflag<1
        x_opt   = c_1;
        v_opt   = fname(c_1)+L_low_1'*(UB-c_1);
    end
    
    LIST_hybBB.x = [LIST_hybBB.x,x_opt];
    fval_up      = min(fval_up,fname(x_opt));
    LIST_hybBB.v = [LIST_hybBB.v,v_opt];
    
    % -> second box 
    LB = inf(X2);
    UB = sup(X2);
    
    if all(L_up_2-L_low_2>sqrt(eps))
        c_2 = (L_up_2.*LB-L_low_2.*UB)./(L_up_2-L_low_2);
    else
        index_n0    = find(L_up_2-L_low_2>sqrt(eps));
        c_2           = 0.5*(LB+UB);
        c_2(index_n0) = (L_up_2(index_n0).*LB(index_n0)-L_low_2(index_n0).*UB(index_n0))./(L_up_2(index_n0)-L_low_2(index_n0));
    end
    
    if 0.125*alpha_relax_2*norm(UB-LB)^2<(L_low_2'*(c_2-UB))
        [x_opt,v_opt,nlpflag] = fmincon(@(x)f_alpha(x,fname,alpha_relax_2,LB,UB),0.5*(LB+UB),[],[],[],[],LB,UB,[],options_nlp);
    else
        nlpflag = 1;
        x_opt   = c_2;
        v_opt   = fname(c_2)+L_low_2'*(UB-c_2);
    end
    
    if nlpflag<1
        x_opt   = c_2;
        v_opt   = fname(c_2)+L_low_2'*(UB-c_2);
    end
    
    LIST_hybBB.x = [LIST_hybBB.x,x_opt];
    fval_up    = min(fval_up,fname(x_opt));
    LIST_hybBB.v = [LIST_hybBB.v,v_opt];
    
    % fathorming step
    index_del                     = find(LIST_hybBB.v>(fval_up+sqrt(eps)));
    LIST_hybBB.box(:,index_del)   = [];
    LIST_hybBB.x(:,index_del)     = [];
    LIST_hybBB.v(index_del)       = [];
    LIST_hybBB.alpha(index_del)   = [];
    LIST_hybBB.L.low(:,index_del) = [];
    LIST_hybBB.L.up(:,index_del)  = [];
    
    % get new box that will be splitted in the next iteration
    [fval_low,index_min] = min(LIST_hybBB.v);
    alpha_relax_old      = LIST_hybBB.alpha(index_min);
    L_low_old            = LIST_hybBB.L.low(:,index_min);
    L_up_old             = LIST_hybBB.L.up(:,index_min);
    X_cur                = LIST_hybBB.box(:,index_min);
    
    if strncmp(split_point,'min',3)
        x_split = LIST_hybBB.x(:,index_min);
    end
    iter = iter+1;
    
    if debug
        disp('upper bound for optimal value')
        fval_up
        disp('lower bound for optimal value')
        fval_low
        disp('smallest and largest alpha in LIST')
        min(LIST_hybBB.alpha)
        max(LIST_hybBB.alpha)
        disp('number of boxes in LIST')
        length(LIST_aBB.box(1,:))
        pause(2)
    end
    
    % check termination criteria
    if fval_up-fval_low<tol_conv
        terminate   = 1;
        exitflag    = 1;
        output.iter = iter;
        output.Term = 'epsilon optimal solution found';
        output.Time = cputime-time;
        X_cur       = infsup(min(inf(X1),inf(X2)),max(sup(X1),sup(X2)));
    elseif iter>iter_max
        terminate   = 1;
        exitflag    = -3;
        output.iter = iter;
        output.Term = 'maximum number of iterations exceeded';
        output.Time = cputime-time;
    elseif max(sup(X_cur)-inf(X_cur))<1e-6
        terminate   = 1;
        exitflag    = -2;
        output.iter = iter;
        output.Term = 'boxes become too small';
        output.Time = cputime-time;
    end
end

if strncmp(options.List,'on',4)
    output.LIST = LIST_hybBB;
end