function [X1,X2] = refine(X,fname,split,split_point,x_split,tolerance)

%**************************************************************************
%*                                                                        *
%* REFINE is the branching function used within the different B&B methods.*
%* The function computes to new boxes arising from splitting a given box  *
%* The way the splitting is performed depends on the selected options     *
%* for Split and Point.                                                   *
%*                                                                        *
%* For more information see:                                              *
%*                                                                        *
%* T.Csendes, D.Ratz                                                      *
%* Subdivision direction selection in interval methods for global         *
%* optimization                                                           *
%* SIAM J. Numer. Anal., 34 (1997), pp. 922-938                           *
%*                                                                        *
%* D.Ratz                                                                 *
%* On the Selection of Subdivision Directions in Interval                 *
%* Branch-and-Bound Methods for Global Optimization                       *
%* Journal of Global Optimization, 7 (1995), pp. 183-207                  *
%*                                                                        *
%**************************************************************************
%*                                                                        *
%*  This program is free software: you can redistribute it and/or modify  *
%*  it under the terms of the GNU General Public License as published by  *
%*  the Free Software Foundation, either version 3 of the License, or     *
%*  (at your option) any later version.                                   *
%*                                                                        *
%*  This program is distributed in the hope that it will be useful,       *
%*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *
%*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
%*  GNU General Public License for more details.                          *
%*                                                                        *
%*  You should have received a copy of the GNU General Public License     *
%*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *
%*                                                                        *
%**************************************************************************
%*                                                                        *
%* P.Steuermann - IOR/Karlsruhe Institute of Technology                   *
%* vers.: 0.0.1-0002 - 27/01/2012                                         *
%*                                                                        *
%**************************************************************************


% compute/set variables and parameters
LB   = inf(X);
UB   = sup(X);
n    = length(LB);
diff = (UB-LB);             
Q    = min(diff)/max(diff); % Q will be used as an indicator 
                            %  to avoid degenerated boxes

if strncmp(split_point,'bary',4) % use barycenter as splitting point
    if Q<tolerance
        split = 'max';
    end
    
    % get the direction along that the box will be split
    % (different rules!)
    if strncmp(split,'max',3)
        [~,index] = max(diff);
    elseif strncmp(split,'gradmax',7)
        [F,DF]    = fname(X);
        [~,index] = max((sup(DF)-inf(DF)).*diff);
    elseif strncmp(split,'gradmid',7)
        [F,DF]    = fname(X);
        w_diff    = DF.*(X-0.5*(LB+UB));
        [~,index] = max(sup(w_diff)-inf(w_diff));
    elseif strncmp(split,'ratz',4)
        [F,DF,D2F] = fname(X);
        w_diff     = (X-0.5*(LB+UB)).*(DF+0.5*sum(D2F'*(X-0.5*(LB+UB)))');
        [~,index]  = max(sup(w_diff)-inf(w_diff));
    elseif strncmp(split,'relmax',6)
        index_zero          = ((LB<=0)&(UB>=0));
        w_diff              = diff;
        w_diff(~index_zero) = diff(~index_zero)./min(abs(LB(~index_zero)),abs(UB(~index_zero)));
        [~,index]           = max(w_diff);
    end
    
    % set new upper/lower bounds for the two new boxes
    x_qs_1        = UB;
    x_qs_2        = LB;
    x_qs_1(index) = (LB(index)+UB(index))*0.5;
    x_qs_2(index) = (LB(index)+UB(index))*0.5;

elseif strncmp(split_point,'min',3) % use minimum of the relaxed function as splitting point
    
    % check if x_split is too close to the boundary of the box and get the
    % indices of the directions where that is the case
    dist_bnd            = union((find((x_split-LB)<=tolerance)),(find((UB-x_split)<=tolerance)));
    index_sea           = 1:n;
    index_sea(dist_bnd) = [];

    % if x_split is near a vertex of the box or if the box could degenerate
    % then the longest edge of the box will be bisected
    if Q<tolerance || length(dist_bnd)==n
        index_sea = 1:n;
        x_split   = 0.5*(LB+UB);
        split     = 'max';
    end
    
    % get the direction along that the box will be split
    % (different rules!)
    if strncmp(split,'max',3)
        [~,index] = max(diff(index_sea));
        index     = index_sea(index);
    elseif strncmp(split,'gradmax',7)
        [F,DF]    = fname(X);
        [~,index] = max((sup(DF(index_sea))-inf(DF(index_sea))).*diff(index_sea));
        index     = index_sea(index);
    elseif strncmp(split,'gradmid',7)
        [F,DF]    = fname(X);
        w_diff    = DF.*(X-0.5*(LB+UB));
        [~,index] = max(sup(w_diff(index_sea))-inf(w_diff(index_sea)));
        index     = index_sea(index);
    elseif strncmp(split,'ratz',4)
        [F,DF,D2F] = fname(X);
        w_diff     = (X-0.5*(LB+UB)).*(DF+0.5*sum(D2F'*(X-0.5*(LB+UB)))');
        [~,index]  = max(sup(w_diff(index_sea))-inf(w_diff(index_sea)));
        index      = index_sea(index);
    elseif strncmp(split,'relmax',6)
        index_zero          = ((LB<=0)&(UB>=0));
        w_diff              = diff;
        w_diff(~index_zero) = diff(~index_zero)./min(abs(LB(~index_zero)),abs(UB(~index_zero)));
        [~,index]           = max(w_diff(index_sea));
        index               = index_sea(index);
    end
    
    % set new upper/lower bounds for the two new boxes
    x_qs_1        = UB;
    x_qs_2        = LB;
    x_qs_1(index) = x_split(index);
    x_qs_2(index) = x_split(index);
end

% set the two new boxes 
X1 = infsup(LB,x_qs_1);
X2 = infsup(x_qs_2,UB);
