function [X_cur,fval_low,fval_up,exitflag,output] = ufBB(fname,lb,ub,constr,options)

%**************************************************************************
%*                                                                        *
%* UFBB_BOX is a branch & bound global optimization solver that uses      *
%* unimodal relaxations of the objective function to solve problems of    *
%* the form:                                                              *
%*                                                                        *
%*    min F(X)  subject to:    C(X) <= 0                                  *
%*     X                     LB <= X <= UB                                *
%*                                                                        *
%* For more information on the unimodal form B&B method see:              *
%*                                                                        *
%* E. Baumann                                                             *
%* Optimal Centered Forms                                                 *
%* BIT Numer. Math., 28 (1988), pp. 80-87                                 *
%*                                                                        *
%* J.Gaerttner                                                            *
%* Unimodale Relaxierungen in der globalen Optimierung: Hybridverafahren  *
%  und numerische Untersuchungen                                          *
%* Bachelorthesis, KIT,                                                   *
%* Institut for Operations Research - Continuous Optimization, 2010.      *
%*                                                                        *
%**************************************************************************
%*                                                                        *
%*  This program is free software: you can redistribute it and/or modify  *
%*  it under the terms of the GNU General Public License as published by  *
%*  the Free Software Foundation, either version 3 of the License, or     *
%*  (at your option) any later version.                                   *
%*                                                                        *
%*  This program is distributed in the hope that it will be useful,       *
%*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *
%*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
%*  GNU General Public License for more details.                          *
%*                                                                        *
%*  You should have received a copy of the GNU General Public License     *
%*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *
%*                                                                        *
%**************************************************************************
%*                                                                        *
%* J.Gaerttner, P.Steuermann - IOR/Karlsruhe Institute of Technology      *
%* vers.: 0.0.1-0001 - 31/01/2012                                         *
%*                                                                        *
%**************************************************************************

debug = 0;
if isfield(options,'debug')
    debug = 1;
end

% set parameters and initialize variables
tol_conv    = options.TolConv;
split       = options.Split;
split_point = options.Point;
iter_max    = options.MaxIter;
LIST_ufBB   = struct;
terminate   = 0;
x_split     = [];
n           = length(lb);
iter        = 1;

time        = cputime;

fval_up     = inf;


if strncmp(options.Relax_para,'one',3)
    paraflag = 0;
elseif strncmp(options.Relax_para,'all',3)
    paraflag = 1;
elseif strncmp(options.Relax_para,'sub',3)
    paraflag = 2;
end

% compute relaxation parameter
X_cur       = infsup(lb,ub);
if paraflag <= 1
    [F,DF] = fname(X_cur);
    L_low  = min(zeros(n,1),inf(DF));
    L_up   = max(zeros(n,1),sup(DF));
else
    [L_low,L_up] = lip_sub_explicit(fname,lb,ub);
    L_low        = min(zeros(n,1),L_low);
    L_up         = max(zeros(n,1),L_up);
end

% compute optimal center/value of the relaxation on [lb,ub]
if all(L_up-L_low>sqrt(eps))
    c = (L_up.*lb-L_low.*ub)./(L_up-L_low);
else
    index_n0    = find(L_up-L_low>sqrt(eps));
    c           = 0.5*(lb+ub);
    c(index_n0) = (L_up(index_n0).*lb(index_n0)-L_low(index_n0).*ub(index_n0))./(L_up(index_n0)-L_low(index_n0));
end

v_opt           = fname(c)+L_low'*(ub-c);
LIST_ufBB.box   = X_cur;
LIST_ufBB.L.low = L_low;
LIST_ufBB.L.up  = L_up;
L_low_old       = L_low;
L_up_old        = L_up;
LIST_ufBB.x     = c;
LIST_ufBB.v     = v_opt;

fval_low        = v_opt;
if all(constr(c)<=0)
    fval_up         = min(fval_up,fname(c));
end

% check termination criteria
if fval_up-fval_low<tol_conv
    terminate   = 1;
    exitflag    = 1;
    output.iter = iter;
    output.Term = 'epsilon optimal solution found';
    output.Time = cputime-time;
elseif iter>=iter_max
    terminate   = 1;
    exitflag    = -3;
    output.iter = iter;
    output.Term = 'maximum number of iterations exceeded';
    output.Time = cputime-time;
end

index_min = 1; % there is only one box that can be split in the first iteration 

% set splitting point for options.Point=='min'
if strncmp(split_point,'min',3)
    x_split = LIST_ufBB.x(:,1);
end

while ~terminate
    % delete it X-cur and all related data from LIST_aBB
    LIST_ufBB.box(:,index_min) = [];
    LIST_ufBB.L.low(:,index_min) = [];
    LIST_ufBB.L.up(:,index_min) = [];
    LIST_ufBB.x(:,index_min)   = [];
    LIST_ufBB.v(index_min)     = [];

    % split current box and add the two new boxes to LIST
    [X1,X2]       = refine(X_cur,fname,split,split_point,x_split,1e-3);
    
    % check if the intersection of the new boxes and the feasible set is
    % not empty - then compute new parameters and compute optimal 
    % centers/values and update LIST_ufBB
    % -> first box
    G = constr(X1);
    if all(inf(G)<=0)
        % compute new parameters L_low, L_up on the new box X1
        if paraflag == 0
            L_low_1 = L_low;
            L_up_1  = L_up;
        elseif paraflag == 1
            [F,DF]  = fname(X1);
            L_low_1 = min(zeros(n,1),inf(DF));
            L_up_1  = max(zeros(n,1),sup(DF));
        else
            [L_low,L_up] = lip_sub_explicit(fname,inf(X1),sup(X1));
            L_low_1      = min(zeros(n,1),L_low);
            L_up_1       = max(zeros(n,1),L_up);
        end

        % rounding errors could occure in the computation of relaxation
        % parameters - we choose the smaller values from the
        % old values and the new values for the relaxation parameters 
        L_low_1 = max(L_low_1,L_low_old);
        L_up_1  = min(L_up_1,L_up_old);
        
        LIST_ufBB.box      = [LIST_ufBB.box,X1];
        LIST_ufBB.L.low    = [LIST_ufBB.L.low,L_low_1];
        LIST_ufBB.L.up     = [LIST_ufBB.L.up,L_up_1];

        LB = inf(X1);
        UB = sup(X1);

        if all(L_up_1-L_low_1>sqrt(eps))
            c_1 = (L_up_1.*LB-L_low_1.*UB)./(L_up_1-L_low_1);
        else
            index_n0      = find(L_up_1-L_low_1>sqrt(eps));
            c_1           = 0.5*(LB+UB);
            c_1(index_n0) = (L_up_1(index_n0).*LB(index_n0)-L_low_1(index_n0).*UB(index_n0))./(L_up_1(index_n0)-L_low_1(index_n0));
        end
        v_opt_1     = fname(c_1)+L_low_1'*(UB-c_1);
        LIST_ufBB.x = [LIST_ufBB.x,c_1];
        LIST_ufBB.v = [LIST_ufBB.v,v_opt_1];
    
        if all(constr(c_1)<=0)
            fval_up     = min([fval_up,fname(c_1)]);
        end
    end
    
    % -> second box
    G = constr(X2);
    if all(inf(G)<=0)
        % compute new parameters L_low, L_up on the new box X2
        if paraflag == 0
            L_low_2 = L_low;
            L_up_2  = L_up;
        elseif paraflag == 1
            [F,DF]  = fname(X2);
            L_low_2 = min(zeros(n,1),inf(DF));
            L_up_2  = max(zeros(n,1),sup(DF));
        else
            [L_low,L_up] = lip_sub_explicit(fname,inf(X2),sup(X2));
            L_low_2      = min(zeros(n,1),L_low);
            L_up_2       = max(zeros(n,1),L_up);
        end

        % rounding errors could occure in the computation of relaxation
        % parameters - we choose the smaller values from the
        % old values and the new values for the relaxation parameters 
        L_low_2 = max(L_low_2,L_low_old);
        L_up_2  = min(L_up_2,L_up_old);

        LIST_ufBB.box      = [LIST_ufBB.box,X2];
        LIST_ufBB.L.low    = [LIST_ufBB.L.low,L_low_2];
        LIST_ufBB.L.up     = [LIST_ufBB.L.up,L_up_2];

        LB = inf(X2);
        UB = sup(X2);

        if all(L_up_2-L_low_2>sqrt(eps))
            c_2 = (L_up_2.*LB-L_low_2.*UB)./(L_up_2-L_low_2);
        else
            index_n0      = find(L_up_2-L_low_2>sqrt(eps));
            c_2           = 0.5*(LB+UB);
            c_2(index_n0) = (L_up_2(index_n0).*LB(index_n0)-L_low_2(index_n0).*UB(index_n0))./(L_up_2(index_n0)-L_low_2(index_n0));
        end
        v_opt_2     = fname(c_2)+L_low_2'*(UB-c_2);
        LIST_ufBB.x = [LIST_ufBB.x,c_2];
        LIST_ufBB.v = [LIST_ufBB.v,v_opt_2];
    
        if all(constr(c_2)<=0)
            fval_up     = min([fval_up,fname(c_2)]);
        end
    end
    
    % fathorming step
    index_del                    = find(LIST_ufBB.v>(fval_up+sqrt(eps)));
    LIST_ufBB.box(:,index_del)   = [];
    LIST_ufBB.x(:,index_del)     = [];
    LIST_ufBB.v(index_del)       = [];
    LIST_ufBB.L.low(:,index_del) = [];
    LIST_ufBB.L.up(:,index_del)  = [];
    
    % get new box that will be splitted in the next iteration
    [fval_low,index_min] = min(LIST_ufBB.v);
    L_low_old            = LIST_ufBB.L.low(:,index_min);
    L_up_old             = LIST_ufBB.L.up(:,index_min);
    X_cur                = LIST_ufBB.box(:,index_min);
    
    if strncmp(split_point,'min',3)
        x_split = LIST_ufBB.x(:,index_min);
    end
    iter = iter+1;
    
    if debug
        disp('upper bound for optimal value')
        fval_up
        disp('lower bound for optimal value')
        fval_low
        disp('number of boxes in LIST')
        length(LIST_ufBB.box(1,:))
        pause(2)
    end
    
    % check termination criteria
    if fval_up-fval_low<tol_conv
        terminate   = 1;
        exitflag    = 1;
        output.iter = iter;
        output.Term = 'epsilon optimal solution found';
        output.Time = cputime-time;
        X_cur       = infsup(min(inf(X1),inf(X2)),max(sup(X1),sup(X2)));
    elseif iter>iter_max
        terminate   = 1;
        exitflag    = -3;
        output.iter = iter;
        output.Term = 'maximum number of iterations exceeded';
        output.Time = cputime-time;
    elseif norm(sup(X_cur)-inf(X_cur))<1e-6
        terminate   = 1;
        exitflag    = -2;
        output.iter = iter;
        output.Term = 'boxes become too small';
        output.Time = cputime-time;
    end
end

if strncmp(options.List,'on',4)
    output.LIST = LIST_ufBB;
end